#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <unistd.h>
#include <dirent.h>
#include <fcntl.h>
#include <string.h>
#include <errno.h>

#define FASTWRITER_SYNCIO_ALIGN 512

#define SYNC_MODE
#define BUFSIZE 2097152
#ifdef SYNC_MODE
# define BLOCK_SIZE 2097152
#else /* SYNC_MODE */
# define BLOCK_SIZE 16384
#endif /* SYNC_MODE */
#define WRITE_INTERVAL 1


int main(int argc, char *argv[]) {
    int err;
    size_t SKIP = 1;
    DIR *dir;
    struct dirent *ent;
    struct timeval start, tv;
    size_t us;
    size_t files = 0;
    size_t total_size = 0;
    size_t last_write = 0;
    size_t skip;
    size_t run;
    ssize_t res;
    size_t max_size = (size_t)-1;
    char *buffer;//[BUFSIZE];
    long double mcoef = 1000000. / (1024 * 1024);
    int flags = O_RDONLY|O_NOATIME|O_LARGEFILE;

#ifdef SYNC_MODE
    flags |= O_DIRECT;
#endif
    
    posix_memalign((void**)&buffer, FASTWRITER_SYNCIO_ALIGN, BUFSIZE);
    
    if (argc < 2) {
	printf("Usage: %s <directory|device> [skip|size]\n", argv[0]);
	exit(0);
    }
    
    if (strstr(argv[1], "/dev/")) {
	if (argc > 2) {
	    max_size = atol(argv[2]);
	    max_size *= 1024 * 1024 * 1024;
	}

	int fd = open(argv[1], flags, 0);
	if (fd < 0) {
	    printf("Unable to open device %s\n", argv[1]);
	    exit(1);
	}
	
	size_t size = BLOCK_SIZE;
	
	gettimeofday(&start, NULL);
        
        res = read(fd, buffer, size);
	while (res > 0) {
	    if (res != size) {
		printf("Incomplete read: %zu bytes read instead of %zu\n", res, size);
		exit(-1);
	    }
	    total_size += res;

	    gettimeofday(&tv, NULL);
	    us = (tv.tv_sec - start.tv_sec) * 1000000 + (tv.tv_usec - start.tv_usec);
	    if ((us - last_write) > WRITE_INTERVAL * 1000000) {
		last_write = us;
		printf("Reading: %s (%lu GB),  Measured speed: %zu MB/s\n", argv[0], total_size / 1024 / 1024 / 1024, (size_t)(mcoef * total_size / us));
	    }
	    
	    if (total_size > max_size) {
		printf("Reading: %s (%lu GB),  Measured speed: %zu MB/s\n", argv[0], total_size / 1024 / 1024 / 1024, (size_t)(mcoef * total_size / us));
		break;
	    }
	
	    res = read(fd, buffer, size);
	}
	
	close(fd);

	if (res < 0) {
	    printf("Read failed with errno %i\n", errno);
	    exit(-1);
	}

	free(buffer);
	
	return 0;
    }

    chdir(argv[1]);

    if (argc > 2) {
	SKIP = atoi(argv[2]);
	
	printf("Skip %zu\n", SKIP);
    }

    gettimeofday(&start, NULL);
    for (run = 0; run < SKIP; run++) {
      skip = 0;
      dir = opendir(".");
      while ((ent = readdir(dir))) {
	struct stat st;
	
	if (((skip++)%SKIP) != run) continue;

	if (stat(ent->d_name, &st)) continue;
	if (!S_ISREG(st.st_mode)) continue;

#ifdef F_MODE
	FILE *f = fopen(ent->d_name, "r");
	if (!f) continue;
#else	
	int fd = open(ent->d_name, flags, 0);
	if (fd < 0) continue;
#endif	
	
	int size = st.st_blksize;
	
	if (size > BUFSIZE) {
	    printf("Buffer too small\n");
	    exit(1);
	}


#ifdef F_MODE
	while (!feof(f)) {
	    err = fread(buffer, 1, size, f);
	    if (err < 0) break;
	}
#else
# ifdef SYNC_MODE
	if (size < BLOCK_SIZE) size = BLOCK_SIZE;
# endif
        err = read(fd, buffer, size);
        while (err > 0) {
	    err = read(fd, buffer, size);
	}
#endif

	if (err < 0) {
	    printf("Read failed\n");
	    exit(1);
	}

#ifdef F_MODE
	fclose(f);
#else
	close(fd);
#endif

	total_size += st.st_size;
	files++;
	
	gettimeofday(&tv, NULL);
	us = (tv.tv_sec - start.tv_sec) * 1000000 + (tv.tv_usec - start.tv_usec);
	printf("Reading: %s (%lu MB), Read: %lu files (%lu GB),  Measured speed: %zu MB/s\n", ent->d_name, st.st_size/1024/1024, files, total_size / 1024 / 1024 / 1024, (size_t)(mcoef * total_size / us));
      }
      closedir(dir);
    }
    
    free(buffer);

}